{==============================================================================|
| Project : Delphree - Synapse                                   | 002.000.002 |
|==============================================================================|
| Content: SNTP client                                                         |
|==============================================================================|
| The contents of this file are subject to the Mozilla Public License Ver. 1.1 |
| (the "License"); you may not use this file except in compliance with the     |
| License. You may obtain a copy of the License at http://www.mozilla.org/MPL/ |
|                                                                              |
| Software distributed under the License is distributed on an "AS IS" basis,   |
| WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for |
| the specific language governing rights and limitations under the License.    |
|==============================================================================|
| The Original Code is Synapse Delphi Library.                                 |
|==============================================================================|
| The Initial Developer of the Original Code is Lukas Gebauer (Czech Republic).|
| Portions created by Lukas Gebauer are Copyright (c)2000,2001.                |
| All Rights Reserved.                                                         |
|==============================================================================|
| Contributor(s):                                                              |
| Change by Patrick Chevalley,  August 29 2001 :                               |
| - Correction for a one second error in DecodeTs.                             |
|   Also change rounding to 1/10000 to avoid error in further calculation.     |
| - New methods :                                                              |
|   - GetNTPhp to take account for roundtrip delay                             |
|    (maybe replace GetNTP by this one)                                        |
|   - SynchroNTP to set the computer time                                      |
|   - EncodeTs to code the NTP timestamp from Tdatetime                        |
|   - GetUTTime to obtain local computer UT time                               |
|   - SetUTtime to set the local computer UT time                              |
| - New Properties :                                                           |
|   - NTPOffset : Local clock offet in seconds                                 |
|   - NTPDelay  : Round trip delay in seconds                                  |
|   - MaxSynchro : Maximum clock offset to correct with SynchroNTP in seconds  |
| - New private values derivated from TNtp.mode used to validate the reply :   |
|   - FLi : warning indicator (leap second warning are ignored)                |
|   - FVn : NTP version                                                        |
|   - Fmode : server mode                                                      |
|==============================================================================|
| History: see HISTORY.HTM from distribution package                           |
|          (Found at URL: http://www.ararat.cz/synapse/)                       |
|==============================================================================}

{$Q-}
{$WEAKPACKAGEUNIT ON}

{$IFNDEF MSWINDOWS}
  {$IFDEF WIN32}
    {$DEFINE MSWINDOWS}
  {$ENDIF}
{$ENDIF}

unit SNTPsend;

interface

uses
  SysUtils,
{$IFDEF MSWINDOWS}
  windows,
{$ENDIF}
{$IFDEF LINUX}
  Libc,
{$ENDIF}
  synsock, blcksock;

const
  cNtpProtocol = 'ntp';

type
  PNtp = ^TNtp;
  TNtp = packed record
    mode: Byte;
    stratum: Byte;
    poll: Byte;
    Precision: Byte;
    RootDelay: Longint;
    RootDisperson: Longint;
    RefID: Longint;
    Ref1: Longint;
    Ref2: Longint;
    Org1: Longint;
    Org2: Longint;
    Rcv1: Longint;
    Rcv2: Longint;
    Xmit1: Longint;
    Xmit2: Longint;
  end;

  TSNTPSend = class(TObject)
  private
    FNTPReply: TNtp;
    FNTPTime: TDateTime;
    FNTPOffset: double;
    FNTPDelay: double;
    FMaxsynchro: double;
    FSntpHost: string;
    FTimeout: Integer;
    FSock: TUDPBlockSocket;
    FBuffer: string;
    FLi, FVn, Fmode : byte;
  public
    constructor Create;
    destructor Destroy; override;
    function DecodeTs(Nsec, Nfrac: Longint): TDateTime;
    procedure EncodeTs(dt: TDateTime ; var Nsec, Nfrac: LongInt) ;
    function GetNTP: Boolean;
    function GetNTPhp: Boolean;
    function SynchroNTP(var timeset : boolean): boolean;
    function GetUTTime : Tdatetime;
    function SetUTTime(newdt: Tdatetime): boolean;
    function GetBroadcastNTP: Boolean;
  published
    property NTPReply: TNtp read FNTPReply;
    property NTPTime: TDateTime read FNTPTime;
    property NTPOffset: Double read FNTPOffset;
    property NTPDelay: Double read FNTPDelay;
    property SntpHost: string read FSntpHost write FSntpHost;
    property Timeout: Integer read FTimeout write FTimeout;
    property Sock: TUDPBlockSocket read FSock;
    property MaxSynchro: double read FMaxSynchro write FMaxSynchro;
  end;

implementation

constructor TSNTPSend.Create;
begin
  inherited Create;
  FSock := TUDPBlockSocket.Create;
  FSock.CreateSocket;
  FTimeout := 5000;
  FMaxSynchro := 3600;
  FSntpHost := cLocalhost;
end;

destructor TSNTPSend.Destroy;
begin
  FSock.Free;
  inherited Destroy;
end;

function TSNTPSend.DecodeTs(Nsec, Nfrac: Longint): TDateTime;
const
  maxi = 4294967295.0;
var
  d, d1: Double;
begin
  Nsec := synsock.htonl(Nsec);
  Nfrac := synsock.htonl(Nfrac);
  d := Nsec;
  if d < 0 then
    d := maxi + d + 1;
  d1 := Nfrac;
  if d1 < 0 then
    d1 := maxi + d1 + 1;
  d1 := d1 / maxi;
  d1 := Trunc(d1 * 10000) / 10000;
  Result := (d + d1) / 86400;
  Result := Result + 2;
end;

procedure TSNTPSend.EncodeTs(dt: TDateTime ; var Nsec, Nfrac: LongInt) ;
const
  maxi = 4294967295.0;
  maxilongint = 2147483647;
var
  d, d1: Double;
begin
  d  := (dt - 2) * 86400;
  d1 := frac(d);
  d  := trunc(d);
  if d>maxilongint then
     d := d - maxi - 1;
  d1 := Trunc(d1 * 10000) / 10000;
  d1 := d1 * maxi;
  if d1>maxilongint then
     d1 := d1 - maxi - 1;
  Nsec:=trunc(d);
  Nfrac:=trunc(d1);
  Nsec := synsock.htonl(Nsec);
  Nfrac := synsock.htonl(Nfrac);
end;

Function TSNTPSend.GetUTTime : Tdatetime;
{$IFDEF MSWINDOWS}
var st : TSystemTime;
begin
 GetSystemTime(st);
 result:=SystemTimeToDateTime(st);
{$ENDIF}
{$IFDEF LINUX}
var
  TV: TTimeVal;
begin
  gettimeofday(TV, nil);
  Result:=UnixDateDelta + (TV.tv_sec+TV.tv_usec / 1000000)/86400;
{$ENDIF}
end;

function TSNTPSend.SetUTTime(newdt: Tdatetime): boolean;
{$IFDEF MSWINDOWS}
var st : TSystemTime;

begin
 DateTimeToSystemTime(newdt,st);
 result:=SetSystemTime(st);
{$ENDIF}
{$IFDEF LINUX}
var
  TV: TTimeVal;
  d : double;
 TZ : Ttimezone;
begin
  result:=false;
  gettimeofday(TV, TZ);
  d:=(newdt-UnixDateDelta)*86400;
  TV.tv_sec:=trunc(d);
  TV.tv_usec:=trunc(frac(d)*1000000);
  if settimeofday(TV, TZ)=-1 then begin
    if ERRNO= EPERM then
       raise exception.create(' Only the super user may set the time!');
  end
  else result:=true;
{$ENDIF}
end;

function TSNTPSend.GetBroadcastNTP: Boolean;
var
  NtpPtr: PNtp;
  x: Integer;
begin
  Result := False;
  FSock.Bind('0.0.0.0', cNtpProtocol);
  if FSock.CanRead(Timeout) then
  begin
    x := FSock.WaitingData;
    SetLength(FBuffer, x);
    FSock.RecvBufferFrom(Pointer(FBuffer), x);
    if (SntpHost = '0.0.0.0') or (FSock.GetRemoteSinIP = SntpHost) then
      if x >= SizeOf(NTPReply) then
      begin
        NtpPtr := Pointer(FBuffer);
        FNTPReply := NtpPtr^;
        FNTPTime := DecodeTs(NTPReply.Xmit1, NTPReply.Xmit2);
        Result := True;
      end;
  end;
end;

function TSNTPSend.GetNTP: Boolean;
var
  q: TNtp;
  NtpPtr: PNtp;
  x: Integer;
begin
  Result := False;
  FSock.Connect(sntphost, cNtpProtocol);
  FillChar(q, SizeOf(q), 0);
  q.mode := $1B;
  FSock.SendBuffer(@q, SizeOf(q));
  if FSock.CanRead(Timeout) then
  begin
    x := FSock.WaitingData;
    SetLength(FBuffer, x);
    FSock.RecvBuffer(Pointer(FBuffer), x);
    if x >= SizeOf(NTPReply) then
    begin
      NtpPtr := Pointer(FBuffer);
      FNTPReply := NtpPtr^;
      FNTPTime := DecodeTs(NTPReply.Xmit1, NTPReply.Xmit2);
      Result := True;
    end;
  end;
end;

function TSNTPSend.GetNTPhp: Boolean;
var
  q: TNtp;
  NtpPtr: PNtp;
  x: Integer;
  t1,t2,t3,t4 : Tdatetime;
begin
  Result := False;
  FSock.Connect(sntphost, cNtpProtocol);
  FillChar(q, SizeOf(q), 0);
  q.mode := $1B;
  t1 := GetUTTime;
  EncodeTs(t1,q.org1,q.org2);
  FSock.SendBuffer(@q, SizeOf(q));
  if FSock.CanRead(Timeout) then
  begin
    t4:=GetUTTime;
    x := FSock.WaitingData;
    SetLength(FBuffer, x);
    FSock.RecvBuffer(Pointer(FBuffer), x);
    if x >= SizeOf(NTPReply) then
    begin
      NtpPtr := Pointer(FBuffer);
      FNTPReply := NtpPtr^;
      FLi:=(NTPReply.mode and $C0) shr 6;
      FVn:=(NTPReply.mode and $38) shr 3;
      Fmode:=NTPReply.mode and $07;
      if (Fli<3) and (Fmode=4) and
         (NTPReply.stratum>=1) and (NTPReply.stratum<=15) and
         (NTPReply.Rcv1<>0) and (NTPReply.Xmit1<>0)
         then begin
           t2 := DecodeTs(NTPReply.Rcv1, NTPReply.Rcv2);
           t3 := DecodeTs(NTPReply.Xmit1, NTPReply.Xmit2);
           FNTPDelay:=((T4 - T1) - (T2 - T3));
           FNTPTime := t3 + FNTPDelay/2;
           FNTPOffset := (((T2 - T1) + (T3 - T4)) / 2)*86400;
           FNTPDelay:=FNTPDelay*86400;
           Result := True;
           end
         else result:=false;
    end;
  end;
end;

function TSNTPSend.SynchroNTP(var timeset : boolean): boolean;
var
  q: TNtp;
  NtpPtr: PNtp;
  x: Integer;
  t1,t2,t3,t4 : Tdatetime;
begin
  Result := False;
  Timeset:= False;
  FSock.Connect(sntphost, cNtpProtocol);
  FillChar(q, SizeOf(q), 0);
  q.mode := $1B;
  t1 := GetUTTime;
  EncodeTs(t1,q.org1,q.org2);
  FSock.SendBuffer(@q, SizeOf(q));
  if FSock.CanRead(Timeout) then
  begin
    t4:=GetUTTime;
    x := FSock.WaitingData;
    SetLength(FBuffer, x);
    FSock.RecvBuffer(Pointer(FBuffer), x);
    if x >= SizeOf(NTPReply) then
    begin
      NtpPtr := Pointer(FBuffer);
      FNTPReply := NtpPtr^;
      FLi:=(NTPReply.mode and $C0) shr 6;
      FVn:=(NTPReply.mode and $38) shr 3;
      Fmode:=NTPReply.mode and $07;
      if (Fli<3) and (Fmode=4) and
         (NTPReply.stratum>=1) and (NTPReply.stratum<=15) and
         (NTPReply.Rcv1<>0) and (NTPReply.Xmit1<>0)
         then begin
           t2 := DecodeTs(NTPReply.Rcv1, NTPReply.Rcv2);
           t3 := DecodeTs(NTPReply.Xmit1, NTPReply.Xmit2);
           FNTPDelay:=((T4 - T1) - (T2 - T3));
           FNTPTime := t3 + FNTPDelay/2;
           FNTPOffset := (((T2 - T1) + (T3 - T4)) / 2)*86400;
           FNTPDelay:=FNTPDelay*86400;
           if (abs(FNTPTime-t1)*86400)<=FMaxSynchro then
              Timeset := SetUTTime(FNTPTime);
           Result := True;
           end
         else result:=false;
    end;
  end;
end;

end.
